/* tvset.c 
 *   by Nathan Laredo (laredo@gnu.org)
 *
 *   with changes by: Gerald Britton <gbritton@mit.edu>
 *   with changes by: Travis Melhiser <melhiser@mythos.org>
 *
 *   See file ChangeLog for recent changes.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <linux/videodev.h>
#include "channels.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/Intrinsic.h>

/* Globals for Xlib interface */
Display *mydisplay;
Window mywindow, myparent;

/* Globals for the X refresh */
Window myrefresh;
void refreshTV(void);
int refreshing = 0;
struct timeval defer_tv;
int defer_refresh = 0;
int defer_min_x = 0;
int defer_min_y = 0;
int defer_max_x = 0;
int defer_max_y = 0;

GC mygc;
XEvent myevent;
int myscreen;
XSizeHints myhint;
KeySym mykey;
XFontStruct *font_struct;
Colormap cmap;
XColor mycolor[256];			/* used for PseudoColor and logo mapping */
unsigned long myforeground, mybackground;
XWindowAttributes attribs;		/* for vis, depth, width, height */
XSetWindowAttributes myattribs;	/* for window creation */
long myevents;					/* keep track of our events */
char title[256];				/* window title */
Pixmap curs_pix;
XColor curs_col;
Cursor mycursor;
int titlestyle = 0;

/* For TV card driving */
struct video_window vwindow;
struct video_window vsavew;
struct video_tuner vtuner;
struct video_picture vpicture;
struct video_buffer vbuffer;
struct video_capability vcap;
struct video_channel vinput;
struct video_audio vaudio;
unsigned char vclip[VIDEO_CLIPMAP_SIZE];	/* need my bttv.c */

int tvfd;
int format = NTSC_BROADCAST;
int channel;
int do_scan = 0;
int do_gray;
int freqtune = 0;
int capture_running = 0;
int audio_running = 1;
int channel_mute = 0;

char tmpbuf[256];				/* temp string space for abuse */

static char *formatname[11] =
{
	"NTSC BROADCAST ", "NTSC CABLE     ", "NTSC JP BRDCAST", "NTSC JP CABLE  ",
	"PAL EUROPE     ", "PAL ITALY      ", "PAL NEW ZEALAND", "PAL AUSTRALIA  ",
	"PAL UHF/IRELAND", "PAL CABLE(B)(G)", "SECAM          "
};

int defer_elapsed(int ms)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (ms < ((tv.tv_sec - defer_tv.tv_sec) * 1000 +
	             ((tv.tv_usec - defer_tv.tv_usec) / 1000)));
}

static void unclip_draw_rectangle(int x, int y, int w, int h)
{
	int i, j;
	/* bitmap is fixed width, 128 bytes (1024 pixels represented) */
	if (x < 0 || y < 0 || w < 0 || h < 0)	/* catch bad clips */
		return;
	/* out of range data should just fall through */
	for (i = y; i < y + h && i < 625; i++)
		for (j = x; j < x + w && j < 1024; j++)
			vclip[(i << 7) + (j >> 3)] &= (~(1 << (j & 7)));
}
/* the following variables are set by get_framebuffer()  */
int *framebuf, bytes_per_line, bits_per_pixel, framewidth, frameheight;

void setformatfreq(int freq)
{
	int i = format;
	ioctl(tvfd, VIDIOCGTUNER, &vtuner);
	if (format <= NTSC_JP_BCAST) {
		vtuner.flags = VIDEO_TUNER_NTSC;
		vtuner.mode = VIDEO_MODE_NTSC;
	} else {
		if (format == 10) {		/* SECAM */
			vtuner.mode = VIDEO_MODE_SECAM;
			vtuner.flags = VIDEO_TUNER_SECAM;
			/* SECAM tuning is wrong, as I saw on my visit to Paris */
			i = PAL_UHF_GHI;	/* fixme!  I have no secam charts! */
		} else {
			vtuner.mode = VIDEO_MODE_PAL;
			vtuner.tuner = VIDEO_TUNER_PAL;
		}
	}
	ioctl(tvfd, VIDIOCSTUNER, &vtuner);
	if (!(vcap.type & VID_TYPE_TUNER))
		return;					/* can't tune */
	freq = tvtuner[freq].freq[i];
	if (!freq)					/* no need to set invalid freq */
		return;
	freq *= 16;
	freq /= 1000;
	freqtune = freq;
	ioctl(tvfd, VIDIOCSFREQ, &freq);
}

void channel_up(void)
{
	int i = format;
	if (format == 10)
		i = PAL_UHF_GHI;
	if ((++channel) == CHAN_ENTRIES)
		channel = 0;
	while (tvtuner[channel].freq[i] == 0) {
		channel++;
		if (channel == CHAN_ENTRIES)
			channel = 0;
	}
	setformatfreq(channel);
}

void channel_down(void)
{
	int i = format;
	if (format == 10)
		i = PAL_UHF_GHI;
	if ((--channel) < 0)
		channel = CHAN_ENTRIES - 1;
	while (tvtuner[channel].freq[i] == 0) {
		channel--;
		if (channel < 0)
			channel = CHAN_ENTRIES - 1;
	}
	setformatfreq(channel);
}


void do_audio(int n)
{
	if (ioctl(tvfd, VIDIOCGAUDIO, &vaudio) == 0) {
		if (n)
			vaudio.flags &= ~VIDEO_AUDIO_MUTE;
		else
			vaudio.flags |= VIDEO_AUDIO_MUTE;
		if (ioctl(tvfd, VIDIOCSAUDIO, &vaudio) < 0)
			perror("set audio");
	}
}

/* this routine is called on clipping/window changes */
void set_window(int apply)
{
	struct video_window vwindow2;
	if (apply) {
		ioctl(tvfd, VIDIOCSWIN, &vwindow);
		ioctl(tvfd, VIDIOCGWIN, &vwindow2);
		vwindow.width = vwindow2.width; vwindow.height = vwindow2.height;
		vwindow.x = vwindow2.x; vwindow.y = vwindow2.y;
		printf("vwindow: %dx%d+%d+%d\n", vwindow.width, vwindow.height,
			vwindow.x, vwindow.y);
		refreshing = 0;
		printf("done refreshing.\n");
	}
}

void do_refresh(void)
{
	struct video_window vwindow2;
	refreshing = 1;
	printf("refreshing...\n");
	memset(vclip, 255, VIDEO_CLIPMAP_SIZE);
	ioctl(tvfd, VIDIOCSWIN, &vwindow);
	ioctl(tvfd, VIDIOCGWIN, &vwindow2);
	vwindow.width = vwindow2.width; vwindow.height = vwindow2.height;
	vwindow.x = vwindow2.x; vwindow.y = vwindow2.y;
	refreshTV();
	//XClearArea(mydisplay, mywindow, 0, 0, 0, 0, True);
	defer_min_x = vwindow.x;
	defer_min_y = vwindow.y;
	defer_max_x = vwindow.x + vwindow.width;
	defer_max_y = vwindow.y + vwindow.height;
}

int setup_tv_device(int n)
{
	sprintf(tmpbuf, "/dev/video%d", n);
	if ((tvfd = open(tmpbuf, O_RDWR)) < 0) {
		perror("open");
		return -1;
	}
	if (ioctl(tvfd, VIDIOCGCAP, &vcap) < 0) {
		perror("get capabilities");
		return -1;
	}
		do_audio(1);
	if (vcap.audios > 0)
		do_audio(1);

	vwindow.clips = (struct video_clip *) vclip;
	/* start with everything clipped, unclip on exposures */
	memset(vclip, 255, VIDEO_CLIPMAP_SIZE);
	vwindow.clipcount = VIDEO_CLIP_BITMAP;
	vwindow.width = 320;
	vwindow.height = 240;
	vwindow.x = vwindow.y = 0;
	ioctl(tvfd, VIDIOCGPICT, &vpicture);
	vpicture.depth = bits_per_pixel;
	switch (bits_per_pixel) {
	case 8:
		vpicture.palette = VIDEO_PALETTE_HI240;
		break;
	case 15:
		vpicture.palette = VIDEO_PALETTE_RAW;
		break;
	case 16:
		vpicture.palette = VIDEO_PALETTE_RGB565;
		break;
	case 24:
		vpicture.palette = VIDEO_PALETTE_RGB24;
		break;
	case 32:
		vpicture.palette = VIDEO_PALETTE_RGB32;
		break;
	default:
		fprintf(stderr, "unsupported depth %d.\n",
				bits_per_pixel);
		return -1;
	}
	if (ioctl(tvfd, VIDIOCSPICT, &vpicture) < 0) {
		perror("set depth");
		fprintf(stderr, "Unable to set %d bit palette\n",
				bits_per_pixel);
		return -1;
	}
	/* these values are set by earlier call to get_framebuffer() */
	vbuffer.width = framewidth;
	vbuffer.height = frameheight;
	vbuffer.base = framebuf;
	if ((vbuffer.depth = bits_per_pixel) == 15)
		vbuffer.depth++;
	vbuffer.bytesperline = bytes_per_line;
	if (ioctl(tvfd, VIDIOCSFBUF, &vbuffer) < 0) {
		perror("set buffer");
		return -1;
	}

	return 0;
}

void tvset_on() { int i = 1; ioctl(tvfd, VIDIOCCAPTURE, &i); }

void close_tv_card()
{
	int i = 0;
	ioctl(tvfd, VIDIOCCAPTURE, &i);
	close(tvfd);
}

int input_quiet = 0;
void set_input(int n)
{
	if (ioctl(tvfd, VIDIOCSCHAN, &n) < 0)
		perror("set channel");
	if (ioctl(tvfd, VIDIOCGCHAN, &vinput) < 0)	/* get name */
		perror("get channel");
	if (n == 0)					/* tuner */
		setformatfreq(channel);
}

/* VideoMode swapping support for full screen tvset */

int full_screen = 0;
XF86VidModeModeInfo **modelines, *fullmode = NULL, *vidmode = NULL;

void vidmode_find(void)
{
	int k, num, clock;
	XF86VidModeModeLine mline;

	/* Save the current video mode so we can get back here */
	XF86VidModeGetModeLine(mydisplay, 0, &clock, &mline);

	/* Find us a nice 640x480 video mode :) */
	XF86VidModeGetAllModeLines(mydisplay, 0, &num, &modelines);
	for (k = 0; k < num; k++) {
		if ((modelines[k]->hdisplay == 640) && (modelines[k]->vdisplay == 480))
			fullmode = modelines[k];
		if ((modelines[k]->hdisplay == mline.hdisplay) &&
			(modelines[k]->vdisplay == mline.vdisplay))
			vidmode = modelines[k];
	}

	{
#if 0
		Window junk;
		XTranslateCoordinates(mydisplay, mywindow,
							  DefaultRootWindow(mydisplay), 0, 0,
							  &initial_x, &initial_y, &junk);
#endif
	}

	if ((fullmode == NULL) || (vidmode == NULL))
		printf("error finding a 640x480 modeline, no fullscreen support\n");
}

void vidmode_640(void)
{
	set_window(1);
	XF86VidModeSwitchToMode(mydisplay, 0, fullmode);
}

void vidmode_def(void)
{
	set_window(1);
	XF86VidModeSwitchToMode(mydisplay, 0, vidmode);
}

void grab_fullscreen(void)
{
	set_window(1);
	vsavew = vwindow;
	XMoveResizeWindow(mydisplay, mywindow, 0, 0, 640, 480);
	vwindow.x = 0;
	vwindow.y = 0;
	vwindow.width = 640;
	vwindow.height = 480;
	vidmode_640();
	XGrabPointer(mydisplay, mywindow, True, 0, GrabModeAsync, GrabModeAsync,
				 mywindow, None, CurrentTime);
	XDefineCursor(mydisplay, mywindow, mycursor);
	XF86VidModeSetViewPort(mydisplay, 0, vwindow.x, vwindow.y);
	XRaiseWindow(mydisplay, mywindow);
	full_screen = 1;
}

void release_fullscreen(void)
{
	set_window(1);
	XUngrabPointer(mydisplay, CurrentTime);
	XUndefineCursor(mydisplay, mywindow);
	XMoveResizeWindow(mydisplay, mywindow, vsavew.x, vsavew.y,
					  vsavew.width, vsavew.height);
	vidmode_def();
	vwindow = vsavew;
	full_screen = 0;
}

/* end full screen support code */

/* the default colour table, for VGA+ colour systems */
int default_red[] =
{0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa,
 0x55, 0xff, 0x55, 0xff, 0x55, 0xff, 0x55, 0xff};
int default_grn[] =
{0x00, 0x00, 0xaa, 0x55, 0x00, 0x00, 0xaa, 0xaa,
 0x55, 0x55, 0xff, 0xff, 0x55, 0x55, 0xff, 0xff};
int default_blu[] =
{0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa,
 0x55, 0x55, 0x55, 0x55, 0xff, 0xff, 0xff, 0xff};

/* Setup pseudocolor color cube or grayscale */
void set_colors(void)
{
	int i, j;
	if ((cmap = XCreateColormap(mydisplay, mywindow, attribs.visual,
								AllocAll)) == 0) {	/* allocate all colors */
		if (!do_gray)
			fprintf(stderr, "unable to allocate colors, using existing map\n");
		return;
	}
	for (i = 0; i < 256; i++) {
		mycolor[i].flags = DoRed | DoGreen | DoBlue;
		mycolor[i].pixel = i;	/* initialize for following Query */
	}
	/* get existing map to preserve some entries if necessary */
	XQueryColors(mydisplay, cmap, mycolor, 256);

	for (i = 0; i < 256; i++)
		if (i < 16 || i > 240 || do_gray) {
			if (do_gray)
				mycolor[i].red = mycolor[i].green = mycolor[i].blue = i * 255;
			else {
				if (i > 239)
					j = i - 240;
				else
					j = i;
				mycolor[i].red = default_red[j] * 255;
				mycolor[i].green = default_grn[j] * 255;
				mycolor[i].blue = default_blu[j] * 255;
			}
		} else {
			j = i - 16;
			if (j > 224)
				break;
			else {
				mycolor[i].red = ((j % 25) / 5) * 16383;
				mycolor[i].green = (j / 25) * 8191;
				mycolor[i].blue = (j % 5) * 16383;
			}
		}
	XStoreColors(mydisplay, cmap, mycolor, 255);
	XSetWindowColormap(mydisplay, mywindow, cmap);
}

/* this routine fills framebuf, bytes_per_line, bits_per_pixel, and do_gray */
/* if 8bpp depth is detected, set_colors() is called */
int get_framebuffer(void)
{
	int major, minor, bank, ram;	/* values we throw away */

	framewidth = attribs.width;
	frameheight = attribs.height;
	if (attribs.depth == 24)
		bits_per_pixel = 32;
	else
		bits_per_pixel = attribs.depth;
	if (bits_per_pixel < 8) {
		fprintf(stderr, "%d bpp is not supported\n",
				bits_per_pixel);
		return -1;
	}
	if (attribs.visual->class == GrayScale ||
		attribs.visual->class == StaticGray)
		do_gray = 1;
	else
		do_gray = 0;

	if (XF86DGAQueryVersion(mydisplay, &major, &minor))
		XF86DGAGetVideoLL(mydisplay, myscreen, (int *) &framebuf,
						  &bytes_per_line, &bank, &ram);
	else {
		fprintf(stderr, "XF86DGA not available.\n");
		fprintf(stderr, "I don't like AccelX.  Use XF86.\n");
		exit(1);
	}

	if (bits_per_pixel == 15)
		bytes_per_line *= 2;
	else
		bytes_per_line *= (bits_per_pixel / 8);
	if (bits_per_pixel == 8)
		set_colors();
	return 0;
}

void update_title(void)
{
	sprintf(title, "tvset [%s] [%dkHz(table=%d)] [%d]",
		tvtuner[channel].name,
		freqtune * 1000 / 16,
		tvtuner[channel].freq[format], vinput.channel);
	XStoreName(mydisplay, mywindow, title);
	XSetIconName(mydisplay, mywindow, title);
}

/* for keyboad entry of channels, c = 0 or 3 chars will enter */
/* On-scren-display of channel as it is entered is done here */

void ch_add_char(char c)
{
	static char cn[5];
	int i, found = -1;

	if (cn[4] != ' ') {			/* initialization flag */
		cn[4] = cn[0] = cn[1] = cn[2] = ' ';
		cn[3] = 0;
	}
	/* rotate new char into channel selector string */
	if (c) {
		cn[0] = cn[1];
		cn[1] = cn[2];
		cn[2] = c;
		cn[3] = 0;
	}
	if (c == 0 || cn[0] != ' ' || cn[1] > '2') {	/* find & change */
		for (i = 0; i < CHAN_ENTRIES && found < 0; i++)
			if (!strcmp(cn, tvtuner[i].name))
				found = i;
		if (found >= 0)
			setformatfreq(channel = found);
		cn[0] = cn[1] = cn[2] = ' ';
		update_title();
		return;
	}
}

int IsAbove(a, b) /* a is above b ? */
{
	int k, nchildren=0, flag=0, above=0;
	Window awin=a, bwin=b, parent, rroot, *children;

	XQueryTree(mydisplay, a, &rroot, &parent, &children, &nchildren);
	while (parent != DefaultRootWindow(mydisplay)) {
		awin = parent;
		XQueryTree(mydisplay, awin, &rroot, &parent, &children, &nchildren);
		if (nchildren) XFree((char *) children);
	}

	XQueryTree(mydisplay, b, &rroot, &parent, &children, &nchildren);
	while (parent != DefaultRootWindow(mydisplay)) {
		bwin = parent;
		XQueryTree(mydisplay, bwin, &rroot, &parent, &children, &nchildren);
		if (nchildren) XFree((char *) children);
	}

	XQueryTree(mydisplay, DefaultRootWindow(mydisplay), &rroot,
	           &parent, &children, &nchildren);
	for (k=0;k<nchildren;k++) {
		if (children[k] == bwin) flag=1;
		if (flag && children[k] == awin) { above=1; break; }
	}
	if (nchildren) XFree((char *) children);
	return above;
}

#define Overlaps(x1, y1, w1, h1, x2, y2, w2, h2) \
	                     (((x1 > x2) ? (x2+w2 > x1) : (x1+w1 > x2)) && \
	                     ((y1 > y2) ? (y2+h2 > y1) : (y1+h1 > y2)))

#define MIN3(a, b, c) ((a < b) ? ((a < c) ? a : c) : ((b < c) ? b : c))
#define MAX3(a, b, c) ((a > b) ? ((a > c) ? a : c) : ((b > c) ? b : c))

int handle_event(void)
{
	Window junk;				/* Scratch Window */

	switch (myevent.type) {
	case VisibilityNotify:
		if (0 && !refreshing && myevent.xvisibility.window == mywindow) {
printf("vis\n");
			defer_refresh = 0;
			do_refresh();
		}
		break;
	case ConfigureNotify:
		if (myevent.xconfigure.window == mywindow) {
			XTranslateCoordinates(mydisplay, mywindow,
								  DefaultRootWindow(mydisplay), 0, 0,
								  &vwindow.x, &vwindow.y, &junk);
			vwindow.width = myevent.xconfigure.width;
			vwindow.height = myevent.xconfigure.height;
			{
				int px, py, pw, ph, pb, pd;
				Window rroot;
				XGetGeometry(mydisplay, myparent, &rroot,
				             &px, &py, &pw, &ph, &pb, &pd);
				defer_min_x = MIN3(vwindow.x, px, defer_min_x);
				defer_min_y = MIN3(vwindow.y, py, defer_min_y);
				defer_max_x = MAX3(vwindow.x + vwindow.width, px + pw,
								   defer_max_x);
				defer_max_y = MAX3(vwindow.y + vwindow.height, py + ph,
								   defer_max_y);
			}
			defer_refresh = 0;
			do_refresh();
		} else { //if (myevent.xconfigure.window != myrefresh) {
			if (!refreshing &&
			Overlaps(vwindow.x, vwindow.y, vwindow.width, vwindow.height,
					 myevent.xconfigure.x, myevent.xconfigure.y,
					 myevent.xconfigure.width, myevent.xconfigure.height) &&
				IsAbove(myevent.xconfigure.window, mywindow)) {
				/* defer an update, we might be opaque dragging ... */
				defer_min_x = MIN3(vwindow.x, myevent.xconfigure.x, defer_min_x);
				defer_min_y = MIN3(vwindow.y, myevent.xconfigure.y, defer_min_y);
				defer_max_x = MAX3(vwindow.x + vwindow.width,
						 myevent.xconfigure.x + myevent.xconfigure.width,
								   defer_max_x);
				defer_max_y = MAX3(vwindow.y + vwindow.height,
						myevent.xconfigure.y + myevent.xconfigure.height,
								   defer_max_y);
				defer_refresh = 1;
				gettimeofday(&defer_tv, NULL);
				//do_refresh();
			}
		}
		break;
	case UnmapNotify:
		if (myevent.xunmap.window == mywindow) {
			defer_refresh = 0;
			do_refresh();
		}
		break;
	case MapNotify:
		if (myevent.xmap.window == mywindow) {
			defer_refresh = 0;
			do_refresh();
		}
		break;
	case DestroyNotify:
		if (myevent.xdestroywindow.window != mywindow)
			break;
		set_window(1);
		return 1;				/* DONE */
		break;
	case Expose:				/* unclip here */
		if (refreshing && myevent.xexpose.window == mywindow) {
			unclip_draw_rectangle(myevent.xexpose.x,
								myevent.xexpose.y, myevent.xexpose.width,
								  myevent.xexpose.height);
			if (refreshing && myevent.xexpose.count == 0)
				set_window(1);
		}
		break;
	case MappingNotify:
		XRefreshKeyboardMapping((XMappingEvent *) & myevent);
		break;
	case KeyPress:
		/* all commands implemented via keypress currently */
		mykey = XKeycodeToKeysym(mydisplay, myevent.xkey.keycode, 0);
		switch (mykey) {
		case XK_q:
			return 1;			/* DONE */
			break;
		case XK_Escape:
			if (full_screen)
				release_fullscreen();
			else
				grab_fullscreen();
			break;
		case XK_KP_Divide:
			if ((vwindow.width * 3) / (vwindow.height * 4) >= 1)
				vwindow.width = (vwindow.height * 4) / 3;
			else
				vwindow.height = (vwindow.width * 3) / 4;
			XResizeWindow(mydisplay, mywindow,
						  vwindow.width, vwindow.height);
//			XTranslateCoordinates(mydisplay, mywindow,
//								  DefaultRootWindow(mydisplay), 0, 0,
//								  &vwindow.x, &vwindow.y, &junk);
//			defer_refresh = 0;
//			do_refresh();
			break;
		case XK_KP_Multiply:
			XResizeWindow(mydisplay, mywindow, 160, 120);
//			XTranslateCoordinates(mydisplay, mywindow,
//								  DefaultRootWindow(mydisplay), 0, 0,
//								  &vwindow.x, &vwindow.y, &junk);
//			vwindow.width = 160;
//			vwindow.height = 120;
//			defer_refresh = 0;
//			do_refresh();
			break;
		case XK_KP_Subtract:
			XResizeWindow(mydisplay, mywindow, 320, 240);
//			XTranslateCoordinates(mydisplay, mywindow,
//								  DefaultRootWindow(mydisplay), 0, 0,
//								  &vwindow.x, &vwindow.y, &junk);
//			vwindow.width = 320;
//			vwindow.height = 240;
//			defer_refresh = 0;
//			do_refresh();
			break;
		case XK_KP_Add:
			XResizeWindow(mydisplay, mywindow, 640, 480);
//			XTranslateCoordinates(mydisplay, mywindow,
//								  DefaultRootWindow(mydisplay), 0, 0,
//								  &vwindow.x, &vwindow.y, &junk);
//			vwindow.width = 640;
//			vwindow.height = 480;
//			defer_refresh = 0;
//			do_refresh();
			break;
		case XK_KP_Decimal:
		case XK_m:
			audio_running = 1 - audio_running;
			do_audio(audio_running);
			break;
		case XK_t:
			if (++titlestyle > 1)
				titlestyle = 0;
			update_title();
			break;
		case XK_f:
			if (++format > 10)
				format = 0;
			setformatfreq(channel);
			update_title();
			break;
		case XK_Next:			/* Page Up */
			update_title();
			break;
		case XK_Prior:			/* Page Dn */
			update_title();
			break;
		case XK_Up:
			if (audio_running) {
				channel_mute = 1;
				do_audio(0);
			}
			channel_up();
			update_title();
			break;
		case XK_Down:
			if (audio_running) {
				channel_mute = 1;
				do_audio(0);
			}
			channel_down();
			update_title();
			break;
		case XK_i:
			if (++vinput.channel >= vcap.channels)
				vinput.channel = 0;
			set_input(vinput.channel);
			update_title();
			break;
		case XK_1:
		case XK_KP_1:
			ch_add_char('1');
			break;
		case XK_2:
		case XK_KP_2:
			ch_add_char('2');
			break;
		case XK_3:
		case XK_KP_3:
			ch_add_char('3');
			break;
		case XK_4:
		case XK_KP_4:
			ch_add_char('4');
			break;
		case XK_5:
		case XK_KP_5:
			ch_add_char('5');
			break;
		case XK_6:
		case XK_KP_6:
			ch_add_char('6');
			break;
		case XK_7:
		case XK_KP_7:
			ch_add_char('7');
			break;
		case XK_8:
		case XK_KP_8:
			ch_add_char('8');
			break;
		case XK_9:
		case XK_KP_9:
			ch_add_char('9');
			break;
		case XK_0:
		case XK_KP_0:
			ch_add_char('0');
			break;
		case XK_a:
			ch_add_char('A');
			break;
		case XK_e:
			ch_add_char('E');
			break;
		case XK_s:
			ch_add_char('S');
			break;
		case XK_ISO_Enter:
		case XK_KP_Enter:
			ch_add_char(0);
			break;
		case XK_Left:
			freqtune--;
			if (freqtune < 0)
				freqtune = 0;
			setformatfreq(channel);
			update_title();
			break;
		case XK_Right:
			freqtune++;
			if (freqtune > 32767)
				freqtune = 32767;
			setformatfreq(channel);
			update_title();
			break;
		case XK_BackSpace:
			do_refresh();
			break;
		}
		break;
	default:
		/* Ignore all other events */
	}							/* switch */
	return 0;
}

int init_tvset(int argc, char **argv)
{
	XGCValues gcvals;
	Window junk, rroot, *children;
	int done, nchildren;

	if (!getenv("DISPLAY")) {
		fprintf(stderr, "DISPLAY not set.\n");
		exit(1);
	}
	if (!(mydisplay = XOpenDisplay(getenv("DISPLAY")))) {
		fprintf(stderr, "Unable to open display.\n");
		exit(1);
	}
	*(vinput.name) = 0;

	XSynchronize(mydisplay, 1);

	XGetWindowAttributes(mydisplay, DefaultRootWindow(mydisplay), &attribs);

	myscreen = DefaultScreen(mydisplay);
	if (attribs.depth != 8) {
		mybackground = BlackPixel(mydisplay, myscreen);
		myforeground = WhitePixel(mydisplay, myscreen);
	} else {					/* pseudocolor / grayscale pixel defs */
		mybackground = 22;
		myforeground = 192;		/* text will be drawn this color */
	}
	/* this sets up reasonable min/max sizes as well as a 1.3-1.5 aspect */
	myhint.width = 720;
	myhint.height = 480;
	myhint.max_width = 910;
	myhint.max_height = 480;
	myhint.min_width = 180;
	myhint.min_width = 120;
	myhint.min_aspect.x = 4;
	myhint.min_aspect.y = 3;
	myhint.max_aspect.x = 15;
	myhint.max_aspect.y = 10;
	/* to add aspect ratio restriction, add to mask below */
	myhint.flags = PSize | PMinSize | PMaxSize /*| PAspect */ ;

	myevents = myattribs.event_mask =
		KeyPressMask |
		KeyReleaseMask |
		ButtonPressMask |
		ButtonReleaseMask |
		ButtonMotionMask |
		ExposureMask |
		VisibilityChangeMask |
		StructureNotifyMask |
		SubstructureNotifyMask;

	myattribs.background_pixel = mybackground;
	mywindow = XCreateWindow(mydisplay, DefaultRootWindow(mydisplay),
		0, 0, myhint.width, myhint.height, 0, attribs.depth, InputOutput,
				  attribs.visual, CWEventMask | CWBackPixel, &myattribs);
	XSetStandardProperties(mydisplay, mywindow, title, title, None, argv,
						   argc, &myhint);
	gcvals.foreground = myforeground;
	gcvals.background = mybackground;
	mygc = XCreateGC(mydisplay, mywindow, GCForeground | GCBackground, &gcvals);
	XMapRaised(mydisplay, mywindow);

	XQueryTree(mydisplay, mywindow, &rroot, &myparent, &children, &nchildren);
	if (myparent == DefaultRootWindow(mydisplay)) myparent = mywindow;

	XSelectInput(mydisplay, DefaultRootWindow(mydisplay),
				 SubstructureNotifyMask);

	/* NOTE: get_framebuffer must succeed before tv card init */
	done = get_framebuffer();	/* fall through to end on failure */
	if (!done)
		done = setup_tv_device(0);	/* fall through on failure */

	XTranslateCoordinates(mydisplay, mywindow,
						  DefaultRootWindow(mydisplay), 0, 0,
						  &vwindow.x, &vwindow.y, &junk);
	vwindow.width = 768;
	vwindow.height = 480;

	font_struct = XLoadQueryFont(mydisplay, "12x24");
	XSetFont(mydisplay, mygc, font_struct->fid);

	/* Scan X VideoModes for full screen support */
	vidmode_find();

	/* Create a 1 pixel cursor to use in full screen mode */
	curs_pix = XCreatePixmap(mydisplay, mywindow, 1, 1, 1);
	curs_col.pixel = 0;
	curs_col.red = 0;
	curs_col.green = 0;
	curs_col.blue = 0;
	curs_col.flags = 0;
	curs_col.pad = 0;
	mycursor = XCreatePixmapCursor(mydisplay, curs_pix, curs_pix, &curs_col,
								   &curs_col, 1, 1);

	return done;
}

void refreshTV(void)
{
	XSetWindowAttributes xswa;
	unsigned long mask;
	static int first = 1;

	xswa.event_mask = ExposureMask | SubstructureNotifyMask;
	XChangeWindowAttributes(mydisplay, mywindow, CWEventMask, &xswa);

	if (first) {
		first = 0;
		xswa.override_redirect = True;
		xswa.backing_store = NotUseful;
		xswa.save_under = False;
		xswa.background_pixel = BlackPixel(mydisplay, myscreen);

		mask = (CWBackPixel | CWSaveUnder | CWBackingStore |
				CWOverrideRedirect);

		myrefresh = XCreateWindow(mydisplay, RootWindow(mydisplay, myscreen),
							 0, 0, vbuffer.width, vbuffer.height, 0,
							 CopyFromParent, InputOutput, CopyFromParent,
							 mask, &xswa);
	}
	XMoveResizeWindow(mydisplay, myrefresh, defer_min_x, defer_min_y,
				   defer_max_x - defer_min_x, defer_max_y - defer_min_y);
	XMapRaised(mydisplay, myrefresh);
	XUnmapWindow(mydisplay, myrefresh);

	xswa.event_mask = myevents;
	XChangeWindowAttributes(mydisplay, mywindow, CWEventMask, &xswa);
}

int main(int argc, char **argv)
{
	int done = 0, retval = 0, xfd;
	fd_set rfds;
	struct timeval tv;

	done = init_tvset(argc, argv);
	update_title();

	defer_min_x = vwindow.x;
	defer_min_y = vwindow.y;
	defer_max_x = vwindow.x + vwindow.width;
	defer_max_y = vwindow.y + vwindow.height;

	set_window(1); tvset_on();				/* Fire up the tube... */

	xfd = ConnectionNumber(mydisplay);

	/* Main event loop */
	while (!done) {
		while (XPending(mydisplay)) { /* process ALL X events */
			XNextEvent(mydisplay, &myevent);
			done = handle_event();
		}
		tv.tv_sec = 0;			/* all timeouts are 1 second multiples */
		tv.tv_usec = 250000;
		FD_ZERO(&rfds);
		FD_SET(xfd, &rfds);
		retval = select(FD_SETSIZE, &rfds, NULL, NULL, &tv);
		if (!retval) { /* timeout */
			if (defer_refresh && defer_elapsed(250)) {
				defer_refresh = 0;
				do_refresh();
			}
			if (channel_mute) {
				do_audio(1);
				channel_mute = 0;
			}
			/* osd expire */
		}
	}	/* while !done */
	if (full_screen)
		release_fullscreen();
	close_tv_card();
	XFreeGC(mydisplay, mygc);
	XDestroyWindow(mydisplay, mywindow);
	XCloseDisplay(mydisplay);
	exit(0);
}								/* main */
